/****************************************************************

	lls.c
		glue layer to go between Proxim's proprietary
		library for their wireless cards and the
		linux networking kernel

	This driver can only be compiled as a module and it
	can only be loaded once per system.  If you feel
	like a linux expert, see README.multicards for notes
	on how to load multiple times.

	This driver does not support probing.  To change irq
	or io port at runtime enter the following:
		insmod rl2mod irq=x io=y
	This won't work for PC Cards, PCI cards, and PnP ISA 
	cards.

	CardType allows you to declare at load time what
	type of card you have.  See the readme for a list
		insmod rl2mod CardType=type

*****************************************************************/

#include <linux/module.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/timer.h>
#include <linux/tqueue.h>
#include <linux/random.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#ifdef 	CONFIG_NET_RADIO  /* If wireless extension exist in the kernel */
#include <linux/wireless.h>
#endif /* CONFIG_NET_RADIO */
#ifdef CONFIG_PCI
#include <linux/pci.h>
#endif /* CONFIG_PCI */
#ifdef USBCARD
#include <linux/usb.h>
#endif /* USBCARD */

/* this wasn't added to the code until later kernels */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) <<8) + (c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <asm/uaccess.h>
#endif

#ifdef PCCARD
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>
#endif
#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <asm/system.h>
#include <pcmcia/ds.h>
#endif /* PCCARD */

#include "version.h"
#include "lld.h"
#include "proxcfg.h"
#include "lls.h"
#include "asm.h"


/****************************************************************
*****************************************************************

General structure:

You'll find this code in 3 sections:
	- the LLS_* callbacks required by the CLLD Library
	- the Prox_* functions registered with the linux kernel
	- the Proxusb_* functions registered with usb
	- the ProxCSS_* functions registered with PC Card

*****************************************************************
*****************************************************************/



/****************************************************************

	LLS Routines Required by the CLLD library
		- LLSReceiveLookAhead
		- LLSSendComplete
		- LLSRegisterInterrupt
		- LLSDeRegisterInterrupt
		- LLSGetCurrentTime
		- LLSGetCurrentTimeLONG
		- LLSRawReceiveLookAhead
		- LLSRawSendComplete
		- LLSSendProximPktComplete
		- LLSPingReceiveLookAhead

	These functions must exist or the library will complain
	with linker errors.

	prototypes declared in lld.h

*****************************************************************/



/****************************************************************
	LLSReceiveLookAhead is called by the library when a 
	packet comes in.  It filters the packet, if necessary,
	requests a buffer from the kernel, and then calls
	LLDReceive to load the buffer.  Also increments
	kernel counters for ethernet stats that are returned
	in Prox_get_stats
*****************************************************************/

void LLSReceiveLookAhead (unsigned_8 *Buffer, unsigned_16 Length, unsigned_16 Status, unsigned_16 RSSI)
{
	struct LLDRxPacketDescriptor rxDesc;
	struct sk_buff *skb;
	struct net_local *lp = thisRangeLan.priv;

#ifdef WIRELESS_EXT
	lp->wstats.qual.level = RSSI;
#endif /* WIRELESS_EXT */

	/* give us a chance to process it before passing up the stack */
	if (IsProximPacket (Buffer))
		LLSProximReceiveLookAhead(Buffer, Length, Status, RSSI);

	/* allocate the buffer for the kernel */
	skb=dev_alloc_skb(Length);
	if(skb==NULL) {
	    	lp->stats.rx_dropped++;
		return;
	}

	/* prepare a 1-item fragment list for LLD */
	rxDesc.LLDRxFragCount = 1;
	rxDesc.LLDRxFragList[0].FSDataLen = Length;
	rxDesc.LLDRxFragList[0].FSDataPtr = skb->data;

	/* retrieve whole packet from LLD */
	LLDReceive(&rxDesc,0,Length);
	
	/* load up the other elements for the kernel */
	skb->len=Length;
	skb->dev=&thisRangeLan;
	skb->protocol=eth_type_trans(skb,&thisRangeLan);
	
	/* increments the stats count */
	lp->stats.rx_packets++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	lp->stats.rx_bytes += Length;
#endif
	if( Buffer[0]&0x01 ) lp->stats.multicast++;
	thisRangeLan.last_rx = jiffies;

	/* tell linux we have a packet*/
	netif_rx(skb);
} /* LLSReceiveLookAhead */


/****************************************************************
	LLSSendComplete is call once for every call to LLDSend
	it is the place where the library declares it is done
	with a packet buffer of that the buffer can be received.
	the buffer construct is to allow the library to be handling
	multiple packets at once.  LLDSend and LLSSendComplete
	don't have to be syncronous.
*****************************************************************/

unsigned_16 LLSSendComplete (struct LLDTxPacketDescriptor *PktDesc)
{
	struct net_local *lp = thisRangeLan.priv;
	int i;

#ifdef BUFFER
	i = PktDesc->LLDTxdriverWS[1];

	if (bfr_list[i].txDesc.LLDTxdriverWS[0] == CLEAR)
		{RL2DEBUG0("buffer wasn't used.  Probably from LLDSendProximPacket.\n");}
	else {
		/* note the successful transmission */
		lp->attempted_tx_packets++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		lp->stats.tx_bytes += bfr_list[i].save_skbuff_ptr->len;
#endif
	/* free up the kernel buffer and the buffer structure element */
		DEV_FREE_SKB(bfr_list[i].save_skbuff_ptr);
		bfr_list[i].txDesc.LLDTxdriverWS[0] = CLEAR;

	/* send completed, free up the resources */
	/* only check for equality, though.  Otherwise, Prox_start_xmit */
	/* can be reentrant, i.e., don't want to set tbusy=0 if _start_xmit */
	/* needs it =1.  */
		if(LLDMAXSEND-1 == outcount--) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
			thisRangeLan.tbusy = CLEAR;
			mark_bh(NET_BH);
#else
			netif_wake_queue(&thisRangeLan);
#endif LINUX_VERSION_CODE
		}
  	}
#else BUFFER
	/* note the successful transmission */
	lp->attempted_tx_packets++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	lp->stats.tx_bytes += saveptr->len;
#endif LINUX_VERSION_CODE
        /* free the SKB and let upper layer know we sent the packet */
        DEV_FREE_SKB(saveptr);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
        thisRangeLan.tbusy = CLEAR;
	mark_bh(NET_BH);
#else
	netif_wake_queue(&thisRangeLan);
#endif LINUX_VERSION_CODE
#endif BUFFER
	return SUCCESS;

} /* LLSSendComplete */


/****************************************************************
	Called only once during LLDinit.  LLDInit is called from Prox_probe.
	return non-zero is the interrupt registration wasn't successful.
	Registers Prox_interrupt
*****************************************************************/

unsigned_8 LLSRegisterInterrupt (unsigned_8 IntNum, void (*CallBackFunc) (void))
{
  if(!LLDUSBCard) {
	if (!LLSPCIFlag) {
	  if(request_irq(IntNum, &Prox_interrupt, 0, thisRangeLan.name, &thisRangeLan))
		  return FAILURE;
	} else {
	  if(request_irq(IntNum, &Prox_interrupt, SA_SHIRQ, thisRangeLan.name, &thisRangeLan))
		  return FAILURE;
	}
  }

	return SUCCESS;
} /* LLSRegisterInterrupt */


/****************************************************************
	Called only if LLSRegisterInterrupt was successful but some later process
	of LLDInit failed.  Should only be called from within LLDInit
*****************************************************************/

unsigned_8 LLSDeRegisterInterrupt (unsigned_8 IntNum)
{
  	if(!LLDUSBCard)
		free_irq(thisRangeLan.irq, &thisRangeLan);
	return SUCCESS;
} /* LLSDeRegisterInterrupt */


/****************************************************************
	Get system time in timer ticks.  The ticker tick delta
	is declared by HZ.  Used only as a delta value.  rollovers
	shouldn't cause a problem.
*****************************************************************/

unsigned_16 LLSGetCurrentTime (void)
{
	/* just system ticks for use by the CLLD. The library
	   assumes HZ =100.  This should fix it if you've changed 
	   HZ for your system  */
	return(jiffies*LIB_HZ/HZ);
} /* LLSGetCurrentTime */


/****************************************************************
	Get system time in timer ticks.
*****************************************************************/

unsigned_32 LLSGetCurrentTimeLONG (void)
{
	/* just system ticks for use by the CLLD. The library
	   assumes HZ =100.  This should fix it if you've changed 
	   HZ for your system  */
	return(jiffies*LIB_HZ/HZ);
} /* LLSGetCurrentTimeLONG */


/****************************************************************
	Similar to LLSReceiveLookAhead but rather than being a data
	packet, this will be a packetized message from the card.
*****************************************************************/

void LLSRawReceiveLookAhead (unsigned_8 Buffer[], unsigned int Length)
{
	/* for processing raw packetized-command responses from the card */
	LLDRawProcess (Buffer, Length);
} /* LLSRawReceiveLookAhead */


/****************************************************************
	Similar to LLSReceiveLookAhead but called when response
	packets are coming back from LLDPing
*****************************************************************/

void LLSPingReceiveLookAhead (unsigned_8 *Buffer, unsigned_16 Length, unsigned_16 Status, unsigned_16 RSSI)
{
	unsigned_16 temp, i, RSSI_percent, plo, phi, p1, p2, h1, h2;
	struct MACList *ptr;

	RL2DEBUG0("%s: PingReceiveLookAhead, Length=%d, Status=%d\n", thisRangeLan.name, Length, Status);

	i = 0;
	/* find where we are in the table */
	while (    (i < (unsigned_16) CurrRSSITable.numval) 
		&& (RSSI < (unsigned_16) CurrRSSITable.val[i].hex)
		&& i < 10
	      ) i++;
	if (RSSI == 0) RSSI_percent = 0xff;
	else if (i == 0) RSSI_percent = 100;
	else if (i >= CurrRSSITable.numval) RSSI_percent = 0;
	else {
	  phi = (unsigned_16) CurrRSSITable.val[0].dbm;
	  plo = (unsigned_16) CurrRSSITable.val[CurrRSSITable.numval-1].dbm;
	  p1 = (unsigned_16) CurrRSSITable.val[i-1].dbm;
	  p2 = (unsigned_16) CurrRSSITable.val[i].dbm;
	  h1 = (unsigned_16) CurrRSSITable.val[i-1].hex;
	  h2 = (unsigned_16) CurrRSSITable.val[i].hex;
	  temp = ((RSSI - h2) * (p2 - p1)) / (h1 - h2);
	  RSSI_percent = ((plo - p2 + temp) * 100) / (plo - phi);
	}

	if(LLSNumMAC < MAXMAC) {
		ptr = kmalloc(sizeof(struct MACList), GFP_KERNEL);
		if (ptr == NULL) {
			printk("%s: kmalloc failed in LLSPingReceiveLookAhead", thisRangeLan.name);
			return;
		}
		for(i=0;i<6;i++)
			ptr->MAC[i] = Buffer[i+6]; 
		ptr->RSSI = (unsigned_8) RSSI_percent; 

		/* should spinlock LLSMACFlag */
		ptr->Prev = TailMACList;
		TailMACList = ptr;
		LLSNumMAC++;
		/* should unspinlock */
	}
} /* LLSPingReceiveLookAhead */


/****************************************************************
	Similar to LLSSendComplete but called as a compliment
	to LLDSendProximPkt
*****************************************************************/

unsigned_16 LLSSendProximPktComplete (struct LLDTxPacketDescriptor FAR *PktDesc, unsigned_8 Status)
{
	RL2DEBUG0("%s: LLSSendProximPktComplete, %d\n", thisRangeLan.name, Status);
	LLSSendComplete (PktDesc);
	return SUCCESS;
} /* LLSSendProximPktComplete */


/****************************************************************
	Similar to LLSSendComplete but called as a compliment 
	to LLDRawSend
*****************************************************************/

unsigned_16 LLSRawSendComplete (unsigned_8 *Buffer)
{
	RL2DEBUG0("%s: RawSendComplete\n", thisRangeLan.name);
	return SUCCESS;
} /* LLSRawSendComplete */


/****************************************************************
	Used primarily to sleep tasks while we wait for a response
	from the card
*****************************************************************/

void LLSSleepTask (unsigned_8 *flag, unsigned_16 timeout)
{
	unsigned long flags;
	LLSWaitFlag = flag;

        save_flags(flags);
        cli();
         
//	sleep_on_timeout(&WaitFlagSet_queue, timeout);
                
        restore_flags(flags);

} /* LLSSleepTask */


/****************************************************************

	Prox_* Linux-required routines start here.  These routines
	will make use of the LLD routines in CLLD and do
	the kernel registration, init, close, etc.

		init_module
		Prox_probe
		Prox_interrupt (a running thread)
		Prox_interrupt_bh

		Prox_open
		Prox_poll  (a running thread)
		Prox_start_xmit (a running thread)

		Prox_close
		cleanup_module

		Prox_ioctl
		Prox_get_stats
		Prox_get_wireless_stats
		Prox_set_multicast_list

	prototypes declared in lls.h

*****************************************************************/


/****************************************************************
	This function is registered with the kernel for device
	io control.  The kernel give the device its registered
	structure, a structure with the information to pass
	between the driver and application, and the device io
	control command.

	Note that when a variable gets changed, it usually doesn't
	take effect until the card is reset.  LLDNeedReset=SET
	will take care of that.

	Also note that this driver doesn't try to read an
	init file.  One should call the configuration
	utility via /etc/conf.module or some other method.

	This function is big.  Probably not worth shortening.
*****************************************************************/

static int  Prox_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
#ifdef WIRELESS_EXT	/* If wireless extension exist in the kernel */
	struct iwreq *	wrq = (struct iwreq *) rq;
#endif /* WIRELESS_EXT */
	struct Prox_ioctl_info *ioc = (struct Prox_ioctl_info *) &rq->ifr_data;
	int i, ret = 0;
	unsigned_8 *ptr;

	/* support proxcfg to be backward compatible */
	if (cmd == SIOCDEVPRIVATE) {
		switch(ioc->cmd) {
		  case CHECK_SIG:
			copy_to_user(ioc->data, PROX_SIG, 6);
			break;
		  case SET_LLDRESET:
			if(!suser())
				return -EPERM;
			LLDNeedReset = SET;
			break;
		  case SET_NODETYPE:
			if(!suser())
				return -EPERM;
			LLDNodeType = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Node type changed: %s.\n", thisRangeLan.name, nodetypes[LLDNodeType]);
			break;
		  case GET_NODETYPE:
		  	*ioc->data = LLDNodeType;
		  	break;
		  case SET_DOMAIN:
			if(!suser())
				return -EPERM;
			LLDDomain = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Domain changed: %d.\n", thisRangeLan.name, LLDDomain);
		    	break;
		  case GET_DOMAIN:
		  	*ioc->data = LLDDomain;
		  	break;
		  case SET_SECID:
			if(!suser())
				return -EPERM;
			printk(KERN_NOTICE "%s: Set Hard Security ID: ", thisRangeLan.name);
			LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
	    		if(LLDSendSetSecurityID(ioc->data))
				printk("failed.\n");
	    		else
				printk("successful.\n");
			LLSNobh = CLEAR;
			break;
		  case SET_SYNC2_SECID:
			if(!suser())
				return -EPERM;
			printk(KERN_NOTICE "%s: Set Hard Security ID: ", thisRangeLan.name);
	    		if(!LLDAsciiSoftSecID[0])
				printk("failed - possibly no security ID learned.\n");
			else {
				LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
	    			if(LLDSendSetSecurityID(LLDAsciiSoftSecID))
					printk("failed.\n");
	    			else
					printk("successful.\n");
				LLSNobh = CLEAR;
			}
			break;
		  case SET_NWID:
			if(!suser())
				return -EPERM;
			printk(KERN_NOTICE "%s: Set Hard Network ID: ", thisRangeLan.name);
			LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
	    		if(LLDSendSetNWID(ioc->data))
				printk("failed.\n");
	    		else
				printk("successful.\n");
			LLSNobh = CLEAR;
			break;
		  case SET_SYNC2_NWID:
			if(!suser())
				return -EPERM;
			printk(KERN_NOTICE "%s: Set Hard Network ID: ", thisRangeLan.name);
			LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
	    		if(LLDSendSetNWID(LLDHRFAsciiSoftSecID))
				printk("failed.\n");
	    		else
				printk("successful.\n");
			LLSNobh = CLEAR;
			break;
		  case SET_SOFTSECID:
			if(!suser())
				return -EPERM;
			if(LLDSoftSecIDNotSupported)
				printk(KERN_NOTICE "%s: Soft Security ID not supported in this card.\n", thisRangeLan.name);
			else {
				printk(KERN_NOTICE "%s: Set Soft Security ID.\n", thisRangeLan.name);
	    			LLDSendSetSoftSecID(ioc->data);
			}
			break;
		  case SET_SOFTNWID:
			if(!suser())
				return -EPERM;
			printk(KERN_NOTICE "%s: Set Soft Network ID.\n", thisRangeLan.name);
	    		LLDSendSetSoftNWID(ioc->data);
			break;
		  case SET_CHANNEL:
			if(!suser())
				return -EPERM;
			LLDChannel = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Channel changed: %d.\n", thisRangeLan.name, LLDChannel);
			break;
		  case GET_CHANNEL:
		  	*ioc->data = LLDChannel;
		  	break;
		  case SET_SUBCHANNEL:
			if(!suser())
				return -EPERM;
			LLDSubChannel = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Subchannel changed: %d.\n", thisRangeLan.name, LLDSubChannel);
			break;
		  case GET_SUBCHANNEL:
		  	*ioc->data = LLDSubChannel;
		  	break;
		  case SET_MASTERNAME:
			if(!suser())
				return -EPERM;
			/* initialize with spaces - 11-byte string padded with spaces */
			strcpy(LLDMSTAName, "           ");
			strcpy(LLDMSTAName, ioc->data);
		    	printk(KERN_NOTICE "%s: Master name changed: %s\n", thisRangeLan.name, LLDMSTAName);
		    	break;
		  case GET_MASTERNAME:
			strcpy(ioc->data, LLDMSTAName);
		  	break;
		  case SET_ROAM_CONFIG:
			if(!suser())
				return -EPERM;
			LLDRoamConfig(LLDRoamConfiguration=(unsigned_8) *ioc->data);
		    	printk(KERN_NOTICE "%s: Roam configuration changed: %s.\n", thisRangeLan.name, roamtypes[LLDRoamConfiguration]);
			break;
		  case GET_ROAM_CONFIG:
		  	*ioc->data = LLDRoamConfiguration;
		  	break;
		  case SET_MAC_OPT:
			if(!suser())
				return -EPERM;
			LLDMacOptimize(LLSMacOptimizeVal=(unsigned_8) *ioc->data);
		    	printk(KERN_NOTICE "%s: MAC Optimize changed: %s.\n", thisRangeLan.name, macconfigtypes[LLSMacOptimizeVal]);
			break;
		  case GET_MAC_OPT:
		  	*ioc->data = LLSMacOptimizeVal;
		  	break;
		  case SET_DEBUG:
			if(!suser())
				return -EPERM;
			rl2_debug |= (unsigned_8) *ioc->data;
			if ( (unsigned_8) *ioc->data ) {
		    		printk(KERN_INFO "%s: Debugging enabled.\n", thisRangeLan.name);
			}
			else {
				rl2_debug = CLEAR;
	    			RL2DEBUG0("%s: This notice doesn't show.\n", thisRangeLan.name);
	    			printk(KERN_INFO "%s: Debugging disabled.\n", thisRangeLan.name);
			}
			break;
		  case GET_DEBUG:
		  	*ioc->data = rl2_debug;
		  	break;
		  case SET_DEBUG_COLOR:
			if(!suser())
				return -EPERM;
			if ( (LLDDebugFlag = (unsigned_8) *ioc->data) ) {
		    		printk(KERN_INFO "%s: Debug color enabled.\n", thisRangeLan.name);
			}
			else {
				printk("\n");
	    			printk(KERN_INFO "%s: Debug color disabled.\n", thisRangeLan.name);
			}
			break;
		  case GET_DEBUG_COLOR:
		  	*ioc->data = LLDDebugFlag;
		  	break;
		  case SET_PEER_TO_PEER:
			if(!suser())
				return -EPERM;
			LLDPeerToPeerFlag = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Peer to Peer: ", thisRangeLan.name);
			if(LLDPeerToPeerFlag) printk("enabled\n");
			else printk("disabled\n");
			break;
		  case GET_PEER_TO_PEER:
		  	*ioc->data = LLDPeerToPeerFlag;
		  	break;
		  case SET_DISABLE_ROAM:
			if(!suser())
				return -EPERM;
			LLDRoamingFlag = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Roaming: ", thisRangeLan.name);
			if(LLDRoamingFlag) printk("disabled\n");
			else printk("enabled\n");
			break;
		  case GET_DISABLE_ROAM:
		  	*ioc->data = LLDRoamingFlag;
		  	break;
		  case SET_ITO:
			if(!suser())
				return -EPERM;
			LLDInactivityTimeOut = ((unsigned_8) *ioc->data)/5;
			LLDTicksToSniff = 0;
			if (!LLDInactivityTimeOut)
				LLDTicksToSniff = ((unsigned_16) *ioc->data) * POLL_FREQ;
		    	printk(KERN_NOTICE "%s: Inactivity Timeout changed: %d seconds.\n", 
		    		thisRangeLan.name, LLDInactivityTimeOut*5 + (LLDTicksToSniff/POLL_FREQ));
			break;
		  case GET_ITO:
		  	*ioc->data = LLDInactivityTimeOut*5 + (LLDTicksToSniff/POLL_FREQ);
		  	break;
		  case GET_SYNCSTATE:
		  	*ioc->data = LLDSyncState;
		  	break;
		  case GET_MSTASYNCNAME:
			strcpy(ioc->data, LLDMSTASyncName);
		  	break;
		  case GET_MSTAADDR:
			copy_to_user(ioc->data, LLDMSTAAddr, 6);
		  	break;
		  case GET_MSTASYNCCHANNEL:
		  	*ioc->data = LLDMSTASyncChannel;
		  	break;
		  case GET_MSTASYNCSUBCHANNEL:
		  	*ioc->data = LLDMSTASyncSubChannel;
		  	break;
		  case GET_SECID:
			if(LLDSoftSecIDNotSupported)
			  copy_to_user(ioc->data, "", 1);
			else
			  copy_to_user(ioc->data, LLDAsciiSoftSecID, strlen(LLDAsciiSoftSecID)+1);
		  	break;
		  case GET_MSTANWID:
			copy_to_user(ioc->data, LLDHRFAsciiSoftSecID, strlen(LLDHRFAsciiSoftSecID)+1);
		  	break;
		  case GET_DRIVERVERSION:
			copy_to_user(ioc->data, CLLDVersion, strlen(CLLDVersion));
			copy_to_user(ioc->data+strlen(CLLDVersion), DriverVersionString, strlen(DriverVersionString)+1);
			break;
		  case GET_ADDR:
			copy_to_user(ioc->data, LLDNodeAddress, 6);
		  	break;
		  case GET_ROMVERSION:
			copy_to_user(ioc->data, LLDROMVersion, strlen(LLDROMVersion)+1);
			if(LLDBootVersion[0]) {
			  copy_to_user(ioc->data + strlen(LLDROMVersion)+1, "/", 1);
			  copy_to_user(ioc->data + strlen(LLDROMVersion)+2, LLDBootVersion, strlen(LLDBootVersion)+1);
			}
			break;
		  case GET_CC:
			LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
		  	if(GetCountryCode()) ret = -EIO;
			LLSNobh = CLEAR;
			ioc->data[0] = LLDCCVal1;
			ioc->data[1] = LLDCCVal2;
		  	break;
		  case GET_DYNCC:
			ioc->data[0] = LLDCurrentCountryCode;
			ioc->data[1] = LLDCurrentFreqOffset;
		  	break;
		  case SET_NODE_OVERRIDE:
			if(!suser())
				return -EPERM;
			if (ioc->data[0] != 0xFF)
			{
				NodeOverrideFlag = SET;
		    		printk(KERN_NOTICE "%s: Node Override: ", thisRangeLan.name);
				ioc->data[0] |= 0x02;
				for(i=0;i<5;i++)
	  			{
		  		  thisRangeLan.dev_addr[i] = LLDNodeAddress[i] = ioc->data[i];
		  		  printk("%2.2x:", thisRangeLan.dev_addr[i]);
		  		}
		  		thisRangeLan.dev_addr[i] = LLDNodeAddress[i] = ioc->data[i];
		  		printk("%2.2x\n", thisRangeLan.dev_addr[i]);
			}
			else
			{
				NodeOverrideFlag = CLEAR;
		    		printk(KERN_NOTICE "%s: Node Override cleared.\n", thisRangeLan.name);
			}
		  	break;
		  case SET_HOP_PERIOD:
			if(!suser())
				return -EPERM;
			LLDHopPeriod = ((unsigned_8) *ioc->data & 0x03) + (LLDHopPeriod & 0xFC);
		    	printk(KERN_NOTICE "%s: Hop Period: %s.\n", thisRangeLan.name, hoptypes[LLDHopPeriod]);
			break;
		  case GET_HOP_PERIOD:
		  	*ioc->data = LLDHopPeriod & 0x03;
		  	break;
		  case SET_BFREQ:
			if(!suser())
				return -EPERM;
			LLDBFreq = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: Beacon Frequency: %d.\n", thisRangeLan.name, LLDBFreq);
			break;
		  case GET_BFREQ:
		  	*ioc->data = LLDBFreq;
		  	break;
		  case SET_BRIDGING:
			if(!suser())
				return -EPERM;
			LLDBridgeFlag = (unsigned_8) *ioc->data;
		    	printk(KERN_NOTICE "%s: bridging: ", thisRangeLan.name);
			if(LLDBridgeFlag) printk("enabled\n");
			else printk("disabled\n");
			break;
		  case GET_BRIDGING:
		  	*ioc->data = LLDBridgeFlag;
		  	break;
		  case GET_PROTO_MODE:
		  	*ioc->data = HomeRF;
		  	break;
		  case GET_CARD_NAME:
		  	*ioc->data = CardType;
		  	break;
		  case GET_CARD_CONFIG:
			/* one or the other is 0 by definition */
		  	*ioc->data = (unsigned_8) LLDCardConfig;
		  	*(ioc->data+1) = (unsigned_8) LLDRL2CardConfig;
		  	break;
		  case SEND_PROX_PING:
			if(!suser())
				return -EPERM;
			/* 0x4, 0x00 means outgoing ping */
			/* ioc->data is the destination MAC */
			/* Buffer length is max 1484 */
		  	LLDSendProximPacket( &LLDProxPkt, Buffer, 1484, ioc->data, 0x4, 0x00);
		  	break;
		  case GET_PROX_PING:
			/* spinlock LLSMACFlag */
			ioc->data[0]=LLSNumMAC;
			for (; LLSNumMAC>0; LLSNumMAC--) {
				for(i=0;i<6;i++)
					ioc->data[((LLSNumMAC-1)*7)+i+1]=TailMACList->MAC[i];
				ioc->data[((LLSNumMAC-1)*7)+i+1]=TailMACList->RSSI;
				ptr = (unsigned_8 *) TailMACList;
				TailMACList = TailMACList->Prev;
				kfree(ptr);
			}
			/* unspinlock LLSMACFlag */
		  	break;
		  case START_AP_SEARCH:
			if(!suser())
				return -EPERM;
		  	LLDMakeAPList(LLDDomain);
		  	break;
		  case STOP_AP_SEARCH:
			if(!suser())
				return -EPERM;
			LLDStopAPSearch(LLDDomain);
			break;
		  case AP_LIST:
			if(!suser())
				return -EPERM;
			LLDPrintAPList((struct MastersList *) ioc->data);
		  	break;
		  case TEACH_MODE:
			if(!suser())
				return -EPERM;
		    	printk(KERN_NOTICE "%s: teach mode initiated: ", thisRangeLan.name);
			LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
			if (InTeachModeFlag)
				printk("failed: already in teach mode.\n");
		  	else if(GotoTeachMode())
				printk("failed.\n");
	    		else
				printk("successful.\n");
			LLSNobh = CLEAR;
		  	break;
		  case LEARN_MODE:
			if(!suser())
				return -EPERM;
		    	printk(KERN_NOTICE "%s: learn mode initiated: ", thisRangeLan.name);
			if (LLDLearningState)
				printk("failed: already in learn mode.\n");
		  	else if (LLDGotoLearnMode())
				printk("failed.\n");
	    		else
				printk("successful.\n");
			break;
		  default:
			ret = -EOPNOTSUPP;
#ifdef PROX_DIAG
		 	ret = Prox_diag_ioctl(dev, ioc->cmd, ioc->data);
#endif /* PROX_DIAG */
		  	break;
		}
	}

#ifdef WIRELESS_EXT	/* If wireless extension exist in the kernel */
	/* support wireless extensions to the kernel */
	else {
		switch(cmd) {
		  /* we'll set a name that identifies the driver */
		  case SIOCGIWNAME:
			strcpy(wrq->u.name, PROX_SIG);
			return 0;
		  /* This domain in RangeLAN2-speak */
		  /* Set it if SU */
		  case SIOCSIWNWID:
		  	if(!suser()) {
				ret = -EPERM;
				break;
			}
			
#if WIRELESS_EXT > 8
			if (wrq->u.nwid.disabled) {
				ret = -EOPNOTSUPP;
				break;
			}
			if (wrq->u.nwid.value<16 && wrq->u.nwid.value>=0) {
				LLDDomain = wrq->u.nwid.value;
				LLDNeedReset = SET;
			}
			else ret = -EINVAL;
#else	/* WIRELESS_EXT */
			if (wrq->u.nwid.on == 0) {
				ret = -EOPNOTSUPP;
				break;
			}
			if (wrq->u.nwid.nwid<16 && wrq->u.nwid.nwid>=0) {
				LLDDomain = wrq->u.nwid.nwid;
				LLDNeedReset = SET;
			}
			else ret = -EINVAL;
#endif /* WIRELESS_EXT */
			break;
		  /* get domain */
		  case SIOCGIWNWID:
#if WIRELESS_EXT > 8
			wrq->u.nwid.value = LLDDomain;
			wrq->u.nwid.disabled = CLEAR;
#else
			wrq->u.nwid.nwid = LLDDomain;
			wrq->u.nwid.on = SET;
#endif
			break;
#if WIRELESS_EXT > 8
    		  case SIOCSIWMODE:
      			switch(wrq->u.mode)
     			{
     			case IW_MODE_MASTER:
       			   LLDNodeType = 2;
       			break;
     			case IW_MODE_INFRA:
       			   LLDNodeType = 0;
       			/* Set the adapter as a station */
       			break;
     			case IW_MODE_SECOND:
       			   LLDNodeType = 1;
     			default:
      			 ret = -EINVAL;
     			}
      		  break;
    		  case SIOCGIWMODE:
      			if(LLDNodeType == 2)
     				wrq->u.mode = IW_MODE_INFRA;
      			else
     			if(LLDNodeType == 1)
       				wrq->u.mode = IW_MODE_MASTER;
     			else
       				wrq->u.mode = IW_MODE_SECOND;
      		  break;
#endif /* WIRELESS_EXT > 8 */
		  /* this is channel in RangeLAN2-speak */
		  /* set it if SU */
		  case SIOCSIWFREQ:
		  	if(!suser())
				return -EPERM;
			if (wrq->u.freq.m<16 && wrq->u.freq.m>0) {
				LLDChannel = (wrq->u.freq.m);
				LLDNeedReset = SET;
			}
			else ret = -EINVAL;
			break;
		  /* get channel */
		  case SIOCGIWFREQ:
			wrq->u.freq.e = 0;
			wrq->u.freq.m = LLDChannel;
			break;
		  case SIOCSIPSUBC:
		  	if(!suser())
				return -EPERM;
			break;
		  case SIOCGIPSUBC:
			break;
		  case SIOCGIWRANGE:
      		  /* Basic checking... */
		      	if(wrq->u.data.pointer != (caddr_t) 0)
			{
			  struct iw_range	range;

			  /* Verify the user buffer */
			  ret = verify_area(VERIFY_WRITE, wrq->u.data.pointer,
					    sizeof(struct iw_range));
			  if(ret)
			    break;

			  /* Set the length (useless : its constant...) */
			  wrq->u.data.length = sizeof(struct iw_range);

			  /* Set information in the range struct */
			  range.throughput = 1.6 * 1024 * 1024;	/* don't argue on this ! */
			  range.min_nwid = 0;
			  range.max_nwid = 15;

			  /* Attempt to recognise 2.00 cards (2.4 GHz frequency selectable) */
			  range.num_channels = 15;
			  range.num_frequency = 0;

			  range.sensitivity = 0;
			  range.max_qual.qual = 2;
			  range.max_qual.level = 255;
			  range.max_qual.noise = 0;

			  /* Copy structure to the user buffer */
			  copy_to_user(wrq->u.data.pointer, &range,
				       sizeof(struct iw_range));
			}
		        break;
    case SIOCGIWPRIV:
      /* Basic checking... */
      if(wrq->u.data.pointer != (caddr_t) 0)
	{
	  struct iw_priv_args	priv[] =
	  {	/* cmd,		set_args,	get_args,	name */
	    { SIOCSIPSUBC, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "setsubchan" },
	    { SIOCGIPSUBC, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "getsubchan" },

	    { SIOCSIPSTNT, IW_PRIV_TYPE_BYTE | 16,	0, "sethisto" },
	    { SIOCGIPSTNT, 0,	    IW_PRIV_TYPE_INT | 16, "gethisto" }
	  };

	  /* Verify the user buffer */
	  ret = verify_area(VERIFY_WRITE, wrq->u.data.pointer,
			    sizeof(priv));
	  if(ret)
	    break;

	  /* Set the number of ioctl available */
	  wrq->u.data.length = 4;

	  /* Copy structure to the user buffer */
	  copy_to_user(wrq->u.data.pointer, (u_char *) priv,
		       sizeof(priv));
	}
      break;
		  default:
			ret = -EOPNOTSUPP;

		}
	}
#endif /* WIRELESS_EXT */

/* make sure we execute on any outstanding changes if the interface
   isn't up. */
	if (!poll_on && !ret)
		LLDPoll();
	return ret;
} /* Prox_ioctl */


/****************************************************************

	Prox_poll ()
	gets called very 100 ms by registering itself as a 
	scheduled function with the kernel.  The
	mechanism by which LLDPoll gets run.  Gets started
	in Prox_open, when the interface is up'd.

	The LLDNeedReset test is because resetting takes a long
	time and gets done in LLDPoll.  busy waits occur.  LLDPoll
	could set LLDNeedReset if something catastrophic happens, 
	like a PC card is pulled out.  By scheduling an extra 1 second
	we allow the kernel time to do other stuff (like update pc card
	status to stop the driver.  Nasty but it works.

*****************************************************************/

void Prox_poll(unsigned long d)
{
	if(poll_on)
	  {
		LLDPoll();
		init_timer(&poll_timer);
		poll_timer.function=Prox_poll;
	     	if (LLDNeedReset)
/*	this extra second is in case of major failure, like pulling out the card.  Constant
	resetting would essentially lock up the system with busy waits */
			poll_timer.expires=jiffies + POLL_FREQ + 1*HZ;
	     	else
			poll_timer.expires=jiffies + POLL_FREQ;
		add_timer(&poll_timer);
	}
} /* Prox_poll */


/****************************************************************
	This is registered with the kernel via
	LLSRegisterInterrupt and is de-registered
	via LLSDeRegisterInterrupt.  It runs LLDIsr, the
	library ISR routine via the bottom half Prox_interrupt_bh.

	A little theory on library's interrupt requirements.  There
	are 4 entry points to the library:  LLDInit, LLDSend, LLDPoll,
	and LLDIsr.  LLDSend, LLDPoll and LLDIsr can "cross"
	each other but then EnableInterrupt and DisableInterrupt
	(in asm.c) but be utilized.  LLDSend and LLDPoll are
	run by linux as task(0), at the same priority.  This means
	the don't cross and won't muck with each other.

	LLDInit is run from Prox_probe which actually runs at
	a lower priority from application space.  Prox_probe
	gets called via init_module which is called from insmod.
	If the card is a PC Card, LLDInit gets called via cardmgr
	from an insertion event.

	LLDIsr, if run as a bottom half will run as task(0) and
	not cross LLDPoll or LLDSend, which is what we want.  

	LLDInit	involves some busy waits as it progresses checking
	each stage of initialization.  Since LLDInit is being
	run at a lower priority than LLDIsr, LLDIsr will be able
	to update flags signalling to LLDInit success or failure
	through the initialization process.

	The one exception to all of this is resetting via LLDPoll.
	If for any reason the card needs to be reset (either
	because you want your new parameters to take effect via
	LLDNeedReset or because the driver thinks the card failed)
	the LLDReset is call inside the library.  Just like for
	LLDInit, LLDReset needs to allow LLDIsr to "cross" it
	to update initialization flags.  To allow this to happen,
	Prox_interrupt checks InResetFlag.  If we are in LLDReset,
	LLDIsr is run from Prox_interrupt at a higher-than-task(0)
	priority.

	Otherwise, LLDIsr is always run as a bottom half.
*****************************************************************/

static void Prox_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = dev_id;
	unsigned_8 status;

     if(!LLDUSBCard) {
	RL2DEBUG0("Entering interrupt top half\n");
	if (dev == NULL  ||  dev->irq != irq)
        {
                printk (KERN_ERR "%s: irq %d for unknown device.\n", dev->name, irq);
                return;
	}

	/* Check if it's for us */
	status = inb(LLDIOAddress1+2) & 0x70;
	if((status != 0x10) && (status != 0x60))
		return;

	/* check for re-entrancy and then call LLDIsr() */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	if(test_and_set_bit(0, (void*)&thisRangeLan.interrupt))
		RL2DEBUG2(KERN_DEBUG "%s: reentrant IRQ\n", thisRangeLan.name);
#endif LINUX_VERSION_CODE

	/* Reset the PCI interrupt */
	if (LLSPCIFlag == SET) {
#ifdef CONFIG_PCI
	  unsigned_16 interruptStatus;
	  interruptStatus = _inpw(LLSIOAddress2+INTERRUPT_REGISTER);
	  if(interruptStatus & INTERRUPT_STATUS) {
		_outpw(LLSIOAddress2+INTERRUPT_REGISTER, interruptStatus | INTERRUPT_CLEAR);
	  }
#endif /* CONFIG_PCI */
	}
      }

	/* process the interrupt */
	/* if reseting, we need LLDIsr to cross LLDReset() or the
	   busy waits will fail */
	if(InResetFlag || LLSNobh)
		Prox_interrupt_bh();
	else {
		queue_task(&LLSISR_Sched, &tq_immediate);
		mark_bh(IMMEDIATE_BH);
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	thisRangeLan.interrupt = CLEAR;
#endif LINUX_VERSION_CODE

	return;
} /* Prox_interrupt */


/****************************************************************
	The ISR bottom half.  Gets scheduled as a task for
	immediate execution by Prox_interrupt.  Prox_interrupt
	is the registered ISR.  See Prox_interrupt for an execption.
	under certain conditions, this will be executed in the
	context of the interrupt.

	If we are a PCI Card, this function needs to reset the
	interrupt.  PC Card interrupt resetting is handled in the
	library.
*****************************************************************/

void Prox_interrupt_bh(void)
{
	LLDIsr();

//	if (LLSWaitFlag && *LLSWaitFlag) wake_up(&WaitFlagSet_queue);
	LLSWaitFlag = NULL;

	return;
} /* Prox_interrupt_bh */


/****************************************************************
	This function is registered with the kernel and
	is called to send ethernet packets via LLDSend.  It works in
	concert with LLSSendComplete.  This function puts
	the packet into the library and LLSSendComplete is
	called by the library when it's done so that the
	packet buffer can be freed.

	If BUFFER is defined, we allow multiple asynchronous
	calls to LLDSend.  If not, the driver is flagged
	as busy until the library finishes handling the packet.

	I like to use BUFFER because the driver could do retries
	to get a packet through.  This could involve queuing and
	packet delays of 100 or more ms.  If we allow the library
	to take more than one packet, the library will process
	packets for other destinations immediately regardless of
	delays for other nodes.
*****************************************************************/

static int Prox_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct net_local *lp = dev->priv;
	int i, result;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	/* check if we've flagged ourselves as busy.  Upper layers
	   shouldn't call us if we're busy unless concerned we are dead. */
	/* LLD will restart itself so we'll ignore upper layer when
			it kicks us */
	if(dev->tbusy) {
		RL2DEBUG0("%s: called when tbusy was set.\n", thisRangeLan.name);
		return -EBUSY;
	}
#endif LINUX_VERSION_CODE

	/* This checks if the system thinks we've missed the tx-done (LLSSendComplete)
	   signal.  If this happens it's either impatient or CLLD has
	   a bug, not sure which. */

	if (skb == NULL ) {
		RL2DEBUG0("%s: called with skb = NULL.\n", thisRangeLan.name);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
		dev_tint(dev);
#endif
 		return 0;
	}

	
	if(skb->len <= 0 || dev == NULL) {
		RL2DEBUG0("%s: null device\n", thisRangeLan.name);
		return 0;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
/* replace with 
        dev->tx_timeout = &my_tx_timeout_func;
        dev->watchdog_timeo = MY_DEV_TX_TIMEOUT;
*/
	/* set us as busy and also check if it was busy.  If
	   calling while we are busy, the kernel is getting
	   impatient.  :)  */
	if(test_and_set_bit(0, (void*)&dev->tbusy)) {
		RL2DEBUG0("%s: Transmitter access conflict.\n", thisRangeLan.name);
		if (skb != NULL) DEV_FREE_SKB(skb);
		return 0;
	}
#else
	netif_stop_queue(dev);
#endif

#ifdef BUFFER
	/* find the first free buffer pointer */
	for (i=0 ; i<LLDMAXSEND ; i++) {
		if ( !(bfr_list[ring].txDesc.LLDTxdriverWS[0]) )
			break;
		if ( ring++ == LLDMAXSEND-1 )
			ring = 0;
	}

	/* none free.  shouldn't get here due to outcount checking.
	   outcount checking tags the driver as busy and won't
	   pass the check above. */
	if (i == LLDMAXSEND) {
		RL2DEBUG0("Called when at LLDMAXSEND\n");
		lp->stats.tx_dropped++;
		DEV_FREE_SKB(skb);
		return 0;
	}

	dev->trans_start = jiffies;
	i = ring;

	/* tag our buffer as used */
	bfr_list[i].txDesc.LLDTxdriverWS[0] = SET;

	/* save the buffer pointer to free later in LLSSendComplete */
	bfr_list[i].save_skbuff_ptr = skb;

	/* fill in the info for transmit decriptor */
	bfr_list[i].txFragList.LLDTxFragList[0].FSDataPtr = skb->data;
	bfr_list[i].txDesc.LLDTxDataLength = 
		bfr_list[i].txFragList.LLDTxFragList[0].FSDataLen = skb->len;

	/* give packet over to LLD for transmission */
	if (bfr_list[i].txDesc.LLDTxdriverWS[0]==CLEAR)
		printk ("Clear going into LLDSend!!!\n");
	result = LLDSend(&bfr_list[i].txDesc);

	if (outcount > LLDMAXSEND-1) {RL2DEBUG0("max outcount exceeded = %d\n", outcount);}
	/* check for max buffering. */
	if (outcount++ < LLDMAXSEND-1) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
		dev->tbusy = CLEAR;
		mark_bh(NET_BH);
#else
		netif_wake_queue(dev);
#endif
	}
	/* we're trying to be very careful here.  LLSSendComplete()
	   is looking for outcount to be = to LLDMAXSEND-1 to set tbusy=CLEAR
	   LLSSendComplete *could* be called by LLDSend so we do the outcount
	   stuff after LLDSend is done */

#else
        /* save the skb for release in LLSSendComplete */
        saveptr = skb;

        /* point frag list to our static structure that has 1 fragments */
        txFragList.LLDTxFragmentCount = 1;
        txDesc.LLDTxFragmentListPtr = &txFragList;
        txFragList.LLDTxFragList[0].FSDataPtr = skb->data;
        
        /* fill in the length for transmit decriptor */
        txDesc.LLDTxDataLength = skb->len;
        txDesc.LLDImmediateDataLength = 0;
        txFragList.LLDTxFragList[0].FSDataLen = skb->len;

        result = LLDSend(&txDesc);
#endif
	
	/* return codes aren't used by LLDSend() as Linux expects but
	   but we'll check anyway.  0 means it took it (not that it was
	   sent).  1 means it was queued (delayed). 0xff means there was
	   an error but I don't belive LLDSend() will return that to us (keep
	   in mind LLDSend() gets call within CLLD also. */
	if(result==1)
	   {RL2DEBUG0("%s: send delayed\n", thisRangeLan.name);}
	else
         if(result) {
            lp->stats.tx_errors++;
            RL2DEBUG0("%s: LLDsend transmit failure!!\n", thisRangeLan.name);
          }

	return 0;
} /* Prox_start_xmit */


/*****************************************************************
	Simply returns a structure to the kernel with
	current ethernet stats.

	We currently only count tx_errors, tx_packets/bytes, rx_dropped
	and rx_packets/bytes.  We don't really have an analog
	for the other stats.
*****************************************************************/

static struct net_device_stats *Prox_get_stats(struct net_device *dev)
{
	struct net_local *lp = dev->priv;
	int clld_dropped, i;
	
	clld_dropped = (int) (LLDOutSyncDrp + LLDTimedQDrp + LLDFailureDrp);
	clld_dropped += (int) (LLDOutSyncQDrp + LLDNoReEntrantDrp + LLDSendDisabledDrp);
	RL2DEBUG1("%d, %d, %d\n", LLDOutSyncDrp, LLDTimedQDrp, LLDFailureDrp);
	RL2DEBUG1("%d, %d, %d\n", LLDOutSyncQDrp, LLDNoReEntrantDrp, LLDSendDisabledDrp);
	RL2DEBUG1("%d, %d\n", LLDNeedResetCnt, LLDSyncState);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	RL2DEBUG1("%d, %ld\n", outcount, thisRangeLan.tbusy);
#endif LINUX_VERSION_CODE
#ifdef BUFFER
        for (i=0; i<LLDMAXSEND; i++) {
	  RL2DEBUG1("%d: %d\n", i, bfr_list[i].txDesc.LLDTxdriverWS[0]);
	}
#endif

	lp->stats.tx_packets = lp->attempted_tx_packets - clld_dropped;
	lp->stats.tx_errors = clld_dropped;
	return &(lp->stats);
} /* Prox_get_stats */


/*****************************************************************
  	Get wireless statistics
  	Called by /proc/net/wireless...
*****************************************************************/

#ifdef WIRELESS_EXT	/* If wireless extension exist in the kernel */
static struct iw_statistics *Prox_get_wireless_stats(struct net_device *	dev)
{
  	struct net_local *lp = dev->priv;

  	if ( (lp->wstats.status = LLDSyncState) )
  		lp->wstats.status = 1;
  	/* Copy data to wireless stuff */
  	if ( (lp->wstats.qual.qual = LLDSyncState) ) {
		if (LLDTxMode) 
			lp->wstats.qual.qual = 2;
		else
			lp->wstats.qual.qual = 1;
  	}
  	lp->wstats.qual.noise = 0;
  	lp->wstats.qual.updated = 0;
  	lp->wstats.discard.nwid = 0L;
  	lp->wstats.discard.code = 0L;
  	lp->wstats.discard.misc = (int) (LLDOutSyncDrp + LLDTimedQDrp + LLDFailureDrp);
  	lp->wstats.discard.misc += (int) (LLDOutSyncQDrp + LLDNoReEntrantDrp + LLDSendDisabledDrp);

  	return &(lp->wstats);
} /* Prox_get_wireless_stats */
#endif	/* WIRELESS_EXT */


/*****************************************************************
	init_module is the driver entry point for loading
	and initializing.  Note that it either calls
	register_netdev or if a PC Card, register_pcmcia_driver.

	register_netdev calls Prox_probe inline which calls
	LLDInit.  LLDInit returns an error (or 0 for success) all the way
	out to init_module.

	register_pcmcia_driver makes ProxCSS_attach available
	for calling later when cardmgr detects a card insertion.  Via
	ProxCSS_attach, LLDInit gets called.  If LLDInit fails, ProxCSS_attach
	has to unload the module.

	In the end, init_modules mostly just checks CardType
	and sets important flags for LLDInit, and set PCI-assigned
	resources if a PCI card.
*****************************************************************/

int init_module(void)
{
	/* Parameters that can be set with 'insmod' */
	thisRangeLan.base_addr = io;
	thisRangeLan.irq = irq;
	LLDWirelessProtocol = wp;

	/* set up for queue LLDIsr () bottom half */
	LLSISR_Sched.routine = (void *) &Prox_interrupt_bh;
	LLSISR_Sched.sync = 0;
	LLSISR_Sched.data = 0;

	/* CardType sets some important flags for the library */
	switch (CardType)
	{
	  case RL2ISA:  /* RangeLAN2 ISA */
		break;
	  case RL2MISA:  /* RangeLAN2 630x series Mini ISA */
		LLDOEM = SET;
		break;
	  case RL2UISA:  /* RangeLAN2 633x series Micro design-in module */
		LLDOEM = SET;
		LLDMicroISA = SET;
		break;
	  case PROXPCCARD:  /* RangeLAN2, Symphony PC, and HomeRF Cards */
		LLDPCMCIA = SET;
		/* NOTE:  rest of flags set in ProxCSS_config() */
		break;
	  case SYMPNPISA:  /* Symphony PnP ISA Card */
		LLDOEM = SET;
		LLDMicroISA = SET;
		break;
	  case PROXPCI:  /* RangeLAN2/Symphony/HomeRF PCI Card */
		LLSPCIFlag = SET;
		/* NOTE:  rest of flags set below according to device number */
		break;
	  case PROXUSB:  /* Symphony/HomeRF USB Card */
		LLDUSBCard = SET;
		break;
	  case PROX633x:  /* 633x Micro Jr Card */
		LLDOEM = SET;
		LLDMicroISA = SET;
		LLDFlag375 = SET;
		break;
	  default:  /* invalid setting */
	  	printk(KERN_WARNING "rlmod.o: Can't init RangeLan, Invalid card type.\n");
		return -ENODEV;
		break;
	}
	
#ifndef USBCARD /* wasn't compiles into the library */
	if (LLDUSBCard) {
		printk(KERN_WARNING "rlmod.o: USB support not compiled into this module.\n");
		return -EINVAL;
	}
#else
	if (!LLDUSBCard) {
		printk(KERN_WARNING "rlmod.o: This module only supports the USB card.\n");
		return -EINVAL;
	}
   	return usb_register(&Proxusb_driver);
#endif /* USBCARD */

#ifndef PCCARD
	if (LLDPCMCIA) {
		printk(KERN_WARNING "rlmod.o: PC Card support not compiled into this module.\n");
		return -EINVAL;
	}
#else
  	if(LLDPCMCIA) {
    		servinfo_t serv;
    		DEBUG(0, "%s\n", versionpc);
    		CardServices(GetCardServicesInfo, &serv);
    		if (serv.Revision != CS_RELEASE_CODE) {
			printk(KERN_NOTICE "rlmod: Card Services release "
			       "does not match!\n");
			return -1;
    		}
   		return register_pcmcia_driver(&dev_info, &ProxCSS_attach, &ProxCSS_detach);
  	}
#endif /* PCCARD */

#ifndef CONFIG_PCI	
	if (LLSPCIFlag) {
		printk(KERN_WARNING "rlmod.o: PCI support not compiled into this module.\n");
		return -EINVAL;
	}
#else
#ifndef PCIBIOS_SUCCESSFUL
#define PCIBIOS_SUCCESSFUL 0x00
#endif /*PCIBIOS_SUCCESSFUL */

	if (LLSPCIFlag == SET) {
	  if (pcibios_present()) {
		unsigned char bus, function;
		int index, result=FAILURE;

		for (index=pci_card_number; 
			index < pci_card_number+PROXIM_MAX_DEV; 
			index++) {
		  if (PCIBIOS_SUCCESSFUL == (result = pcibios_find_device(PCI_VENDOR_ID_PROXIM, PCI_DEVICE_ID_PROXIM, index, &bus, &function)))
		  {
			LLDOEM = SET;
			LLDMicroISA = SET;
			LLDFlag375 = CLEAR;
		  } else if (PCIBIOS_SUCCESSFUL == (result = pcibios_find_device(PCI_VENDOR_ID_PROXIM, PCI_DEVICE_ID_PROXIM_NONE, index, &bus, &function)))
		  {
			LLDOEM = SET;
			LLDMicroISA = SET;
			LLDFlag375 = SET;
		  }
		  if (result != PCIBIOS_SUCCESSFUL)
		    break;
	/* found a card */		    
                        {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
                                struct pci_dev *pdev = pci_find_slot(bus, function);
				LLSIOAddress2 = pdev->resource[1].start;
				thisRangeLan.base_addr = pdev->resource[3].start;
				LLDIORange1 = 16;
				thisRangeLan.irq = pdev->irq;
				LLSPCIDevice = pdev->device;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,55)
                                struct pci_dev *pdev = pci_find_slot(bus, function);
                                LLSIOAddress2 = pdev->base_address[1] & PCI_BASE_ADDRESS_IO_MASK;
                                thisRangeLan.base_addr = pdev->base_address[3] & PCI_BASE_ADDRESS_IO_MASK;
			  	LLDIORange1 = 16;  /* PCI minimum */
                                thisRangeLan.irq = pdev->irq;
				LLSPCIDevice = pdev->device;
#else /* LINUX_VERSION_CODE */
                                u32 pci_ioaddr;
                                u8 pci_irq_line;
                                pcibios_read_config_byte(bus, function, PCI_INTERRUPT_LINE, &pci_irq_line);
                                pcibios_read_config_dword(bus, function, PCI_BASE_ADDRESS_3, &pci_ioaddr);
                                thisRangeLan.base_addr = pci_ioaddr & ~3;
			  	LLDIORange1 = 16;  /* PCI minimum */
                                pcibios_read_config_dword(bus, function, PCI_BASE_ADDRESS_1, &pci_ioaddr);
                                LLSIOAddress2 = pci_ioaddr & ~3;
                                thisRangeLan.irq = pci_irq_line;
				LLSPCIDevice = pdev->device;
#endif /* LINUX_VERSION_CODE */
                        }
		}
		if (result != PCIBIOS_SUCCESSFUL) {
		  printk(KERN_WARNING "rlmod.o pci error: %d\n", result);
		  return -ENODEV;
		}
		if (index == 0)
		  return -ENODEV;
	  }
	  else
		return -ENODEV;
	}
#endif /* CONFIG_PCI */

	/* Register the device with the kernel.  Note that thisRangeLan has */
	/* Prox_probe as a paramter.  register_netdev calls it.  Prox_probe */
	/* does the initialization stuff. */
	return register_netdev(&thisRangeLan);

} /* init_module */


/****************************************************************
   Init Failure (reason) Codes for Prox_probe(), i.e, LLDInit()
   0       -       Successful Initialization
   1       -       Error resetting the RFNC
   2       -       Error setting the interrupt vector
   3       -       Error sending the initialize command
   4       -       Error getting the ROM version
   5       -       Error getting the global address
   6       -       Error setting the local address
   7       -       Error setting the inactivity time out
   8       -       Error unknown
   9       -       Error setting security ID
   10      -       Error sending configure MAC
   11      -       Error sending set roaming parameters command
   12      -       Error sending multicast command
   13      -       Error in card and socket services
   14      -       Error configuring socket - improper card seating
   15      -       Error configuring socket - power imporperly supplied
   16      -       Error configuring socket - no PCMCIA bus ready signal

	Prox_probe is called by init_module and is run at load time.
	See init_module comments for more info.
****************************************************************/

static int Prox_probe(struct net_device *dev)
{
	int inival,i;
	unsigned_32 rndval;
	
	/* If it's a PC Card, card services already got the window for 
	   us.  Just use it. */
	if (LLDPCMCIA) {
	  dev->base_addr = LLDIOAddress1;
	  dev->irq = LLDIntLine1; /* irq gets registered from LLDInit() */
	}
	else if (!LLDUSBCard)
	{
	  /* this should be code to probe for cards */
	  /* We don't support it and supposedly it's bad to */
	  /* autoprobe in a module. */
	  if(dev->base_addr == 0)
	  {
		printk(KERN_WARNING "%s: RangeLan can't autoprobe\n",dev->name);
		return -ENODEV;
	  }

	  /* check if this io region is used. */
	  if (check_region(dev->base_addr, LLDIORange1))
		return -ENODEV;
	  /* request the resources and register the region */
	  request_region(dev->base_addr, LLDIORange1, thisRangeLan.name);
	  LLDIOAddress1 = dev->base_addr;

	  /* register PCI region */
	  if (LLSPCIFlag == SET) {
	    if (check_region(LLSIOAddress2, LLSIORange2)) {
		release_region(dev->base_addr, LLDIORange1);
		return -ENODEV;
	    }
	    /* request the resources and register the region */
	    request_region(LLSIOAddress2, LLSIORange2, thisRangeLan.name);
	  }

	  LLDIntLine1 = dev->irq;
	}

	/* randomize channel and subchannel.  This is for Symphony where you
	   don't know how your next door neighbor set up theirs.  Override to
	   your desired value with ioctl */
	get_random_bytes(&rndval, 4);
	LLDChannel = (rndval%15)+1;
	get_random_bytes(&rndval, 4);
	LLDSubChannel = (rndval%15)+1;

	/* Set the driver version */
	strcpy(DriverVersionString, RL2VERSION);
	printk(KERN_INFO "%s", version);

	/* LLDInit() brings up the card with all default values or values
	   that are preset.  It returns error codes as commented above.  */
	LLSNobh = SET; /* keeps ISR out of bottom half to allow crossing */
	inival=LLDInit();

	if(inival == FIRMWARE_ERR)
		printk(KERN_INFO "rlmod.o: Firmware error.  Reload firmware.\n");
	else if(inival) {
	  /* in case of spurious interrupts during a fail to init.  Can
	     have a subtle problem with interrupts being masked and then
	     unregistered.  irq is then lost to the OS. */
	  printk(KERN_WARNING "%s: %s.\n",dev->name,error_codes[inival]);
	  if (!LLDUSBCard) release_region(dev->base_addr, LLDIORange1);
	  if (LLSPCIFlag == SET) release_region(LLSIOAddress2, LLSIORange2);
	  return -ENODEV;
	}
	LLSNobh = CLEAR;

	/* print info to logs since we initialized properly */
	printk(KERN_INFO "rlmod.o: Linux port by Paul Chinn (loomer@1000klub.com).\n");
#ifdef BUFFER
	printk(KERN_INFO "%s: %s, buffers enabled, ",
#else
	printk("%s: %s, no buffers, ",
#endif
	  	dev->name, card_types[CardType]);
	printk("fw: %s.\n", LLDROMVersion);
	printk(KERN_INFO "%s: ", dev->name);
	for(i=0;i<5;i++)
	  {
		  dev->dev_addr[i] = LLDNodeAddress[i];
		  printk("%2.2x:", dev->dev_addr[i]);
		  }
		  dev->dev_addr[i] = LLDNodeAddress[i];
		  printk("%2.2x, ", dev->dev_addr[i]);
	printk("IO: %#3.3lx-%x, ", dev->base_addr, LLDIORange1-1);
	if(LLSPCIFlag==SET) printk("extra PCI IO: %#3.3x, ", LLSIOAddress2);
	printk("irq %d.\n", dev->irq);
	if (LLDPCMCIA) printk(KERN_INFO "%s: Parts from dummy_cs.c 1.9 1998/07/16 23:44:06 (David Hinds).\n", dev->name);

	/* get the RSSI calibration table */
	LLSNobh = SET; /* makes LLDIsr *not* run as a bottom half */
	GetRSSITable();
	LLSNobh = CLEAR;

	/* Initialize the private space in the device structure. */
	if (dev->priv == NULL) {
		dev->priv = kmalloc(sizeof(struct net_local), GFP_KERNEL);
		if (dev->priv == NULL)
			return -ENOMEM;
	}
	memset(dev->priv, 0, sizeof(struct net_local));

	/* register some routines with the kernel */
	dev->open = &Prox_open;
	dev->stop = &Prox_close;
	dev->hard_start_xmit = &Prox_start_xmit;
	dev->get_stats = Prox_get_stats;
	dev->set_multicast_list = &Prox_set_multicast_list;
	dev->do_ioctl = Prox_ioctl;
#ifdef WIRELESS_EXT	/* If wireless extension exist in the kernel */
	dev->get_wireless_stats = Prox_get_wireless_stats;
	printk(KERN_INFO "%s: Wireless extensions enabled\n",dev->name);
#endif

	ether_setup(dev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	netif_start_queue(dev);
#endif LINUX_VERSION_CODE

	/* all done */
	return 0;
} /* Prox_probe */


/****************************************************************
	called by ifconfig when interface is up'd.  Turns on
	Prox_poll, initializes the send buffers and stats,
	functions to be called by kernel.  Increments mod use
	counter.

	if LLDPCMCIA, also reset the card.  We don't know if
	the power was turned off so we need to be sure
	it's initialized.  Otherwise, we can assume it is.
****************************************************************/

static int Prox_open(struct net_device *dev)
{
	struct net_local *lp = dev->priv;
	int i;

	/* start the polling timer */
	poll_on = 1;
	init_timer(&poll_timer);
	poll_timer.expires=POLL_FREQ;
	poll_timer.function=Prox_poll;
	add_timer(&poll_timer);

#ifdef BUFFER
	outcount = ring = 0;
	/* initialize our pre-defined buffers for CLLD */
	for (i=0 ; i<LLDMAXSEND ; i++) {
		bfr_list[i].txDesc.LLDTxdriverWS[0] = CLEAR;
		bfr_list[i].txDesc.LLDTxdriverWS[1] = i;
		bfr_list[i].txFragList.LLDTxFragmentCount = 1;
		bfr_list[i].txDesc.LLDTxFragmentListPtr = &bfr_list[i].txFragList;
		bfr_list[i].txDesc.LLDImmediateDataLength = 0;
	}
#endif /* BUFFER */

	/* initialize some stats */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	lp->stats.rx_bytes = lp->stats.tx_bytes = 0;
#endif
	lp->stats.rx_packets = lp->attempted_tx_packets = 0;
    	lp->stats.rx_dropped = lp->stats.tx_dropped = 0;
    	lp->stats.rx_errors = lp->stats.tx_errors = 0;
	lp->stats.multicast = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	/* init some stuff in our kernel device structure */
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;
#endif LINUX_VERSION_CODE

	/* flag the module as 'in use' */
	MOD_INC_USE_COUNT;

	/* reset the card to be sure it's active */
	if (LLDPCMCIA) LLDNeedReset = SET;

	return 0;
} /* Prox_open */


/****************************************************************
	cleanup_module is called via rmmod to unload the
	module.

	For pcmcia, called via ProxCSS_release
****************************************************************/

void cleanup_module(void)
{
#ifdef PCCARD
  	if(LLDPCMCIA) {
    	  DEBUG(0, "dummy_cs: unloading\n");
    	  unregister_pcmcia_driver(&dev_info);
  	}
#endif /* PCCARD */

	/* PCMCIA handles registering and unregistering IO */
	if(!LLDPCMCIA) {
		if (check_region(LLDIOAddress1, LLDIORange1))
			release_region(LLDIOAddress1, LLDIORange1);
		if (LLSIOAddress2) /* check region doesn't work right if dma is a 0 */
			release_region(LLSIOAddress2, LLSIORange2);
	}

	/* free kernel memory that isn't freed otherwise.
	   if it is used, we presume that we have a
	   loaded/initialized driver.  Otherwise,
	   skip the rest */
	if (thisRangeLan.priv) {
		kfree(thisRangeLan.priv);
		if(!LLDUSBCard) {
			free_irq(thisRangeLan.irq, &thisRangeLan);
			/* unregister the driver from the kernel */
			unregister_netdev(&thisRangeLan);
		}
	}

#ifdef USBCARD
	if(LLDUSBCard)
		usb_deregister(&Proxusb_driver);
#endif /* USBCARD */

} /* cleanup_module */


/****************************************************************
	called by ifconfig down.  This stops the card
	and ensures it won't generate interrupts.  Also,
	LLDPoll is stopped and we shouldn't get calls to
	LLDSend (Prox_start_xmit).

	To start again, call Prox_open
****************************************************************/

static int Prox_close(struct net_device *dev)
{
	/* flag the module as 'not in use' */
	MOD_DEC_USE_COUNT;

	/* shut down the poll timer */
	del_timer(&poll_timer);
	poll_on = 0;

	/* stop the card. */
	if (!LLDPCMCIA || PCMCIACardInserted) LLDStop();

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	/* tell the kernel the device isn't running */
	dev->start = 0;
#else
	netif_stop_queue(dev);
#endif LINUX_VERSION_CODE

	return 0;
} /* Prox_close */


/****************************************************************
	Wireless cards can't really be promiscuous.  Wireless
	uses ACK's because you can't count on RF like you can
	a wire.  LLDBridgeFlag provides for promiscuous
	operation as a bridge.  This is truely promiscuous
	like you'd expect for a sniffer.
*****************************************************************/

static void Prox_set_multicast_list(struct net_device *	dev)
{
  	if(dev->flags & IFF_PROMISC) LLDBridgeFlag = SET;
  	else if( !(dev->flags & IFF_PROMISC) ) LLDBridgeFlag = CLEAR;

	/* turn on reception of multicast - only applies to
	   HomeRF mode.  OpenAir will always receive multicast */
	
  	if(dev->flags & IFF_MULTICAST) 
	{
		LLSNobh = SET;
		LLDMulticast(1);
		LLSNobh = CLEAR;
	}
  	else if( !(dev->flags & IFF_MULTICAST) ) 
	{
		LLSNobh = SET;
		LLDMulticast(0);
		LLSNobh = CLEAR;
	}
}


/****************************************************************

	Proxusb_* usb Linux-required functions start here.

		Proxusb_probe
		Proxusb_close

*****************************************************************/

#ifdef USBCARD
struct prxm {
  struct usb_device    *usb;
  struct net_device    *net;
};

struct usb_eth_dev {
	char	*name;
	__u16	vendor;
	__u16	device;
	void	*private;
};

static struct usb_eth_dev usb_dev_id[] = {
	{"INTEL AnyPoint HomeRF", 0x8086, 0x0001, NULL},
	{"Proxim Symphony II", 0x08c4, 0x1701, NULL},
	{NULL, 0, 0, NULL}
};

static int check_device_ids( __u16 vendor, __u16 product )
{
	int i=0;
	
	while ( usb_dev_id[i].name ) {
		if ( (usb_dev_id[i].vendor == vendor) && 
		     (usb_dev_id[i].device == product) ) {
		  RL2DEBUG6("Found VID 0x%x PID 0x%x\n",
			 usb_dev_id[i].vendor, usb_dev_id[i].device);
		  return i;
		}
		i++;
	}
	return	-1;
}

/****************************************************************

	called from init_module for usb initialization via
	usb_register().

*****************************************************************/

static void *Proxusb_probe (struct usb_device *dev, unsigned int ifnum)
{
	int dev_indx;
	struct prxm *prxm;

	/* set up descriptors and end points */

	if ( (dev_indx = check_device_ids(dev->descriptor.idVendor, dev->descriptor.idProduct)) == -1 ) {
		return NULL;
	}
	RL2DEBUG6("proxim probe: vid 0x%x pid 0x%x ifnum %d\n",
	       dev->descriptor.idVendor, dev->descriptor.idProduct, ifnum);

	if (usb_set_configuration(dev, dev->config[0].bConfigurationValue)) {
		err("usb_set_configuration() failed");
		return NULL;
	}
	RL2DEBUG6("proxim probe: set config\n");
	if(!((void *)prxm = kmalloc(sizeof(struct prxm), GFP_KERNEL))) {
		err("out of memory allocating device structure");
		return NULL;
	}
	memset((void *)prxm, 0, sizeof(struct prxm));

	RL2DEBUG6("Allocating prxm struct\n");
	prxm->usb = dev;
	prxm->net = &thisRangeLan;

	/* check card capability before setting ... */
	LLDFlag375 = SET;

	DevExtInit(prxm->usb, prxm->net);

	register_netdev(prxm->net); /* calls Prox_probe() */
	return prxm;
}


/****************************************************************

	called from usb via usb_resgister().  That is registered
	in init_module();

*****************************************************************/

static void Proxusb_close (struct usb_device *dev, void *ptr)
{
	struct prxm *prxm = ptr;

	if (!prxm) {
	  warn("unregistering non-existant device");
	  return;
	}

	unregister_netdev(prxm->net);

	DevExtFree();

	kfree(prxm);
}
#endif /* USBCARD */


/****************************************************************

	ProxCSS_* Linux-required routines start here.  These routines
	will make use of the Prox_* and handle pcmcia functionality

		ProxCSS_cs_error
		ProxCSS_attach
		ProxCSS_detach
		ProxCSS_config
		ProxCSS_release
		ProxCSS_event

*****************************************************************/


#ifdef PCCARD
/****************************************************************

    The initial developer of the original dummy_cs code is David A. Hinds
    <dhinds@hyper.stanford.edu>.  Portions created by David A. Hinds
    are Copyright (C) 1998 David A. Hinds.  All Rights Reserved.
    
*****************************************************************/


/****************************************************************
	Handles error reporting
*****************************************************************/

static void ProxCSS_cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}


/****************************************************************

    ProxCSS_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.

    This gets called by register_pcmcia_driver
    
*****************************************************************/

static dev_link_t *ProxCSS_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    local_info_t *local;
    int ret, i;
    
    DEBUG(0, "rlmod_attach()\n");

    /* Initialize the dev_link_t structure */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &ProxCSS_release;
    link->release.data = (u_long)link;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    if (irq_list[0] == -1)
	link->irq.IRQInfo2 = irq_mask;
    else
	for (i = 0; i < 4; i++)
	    link->irq.IRQInfo2 |= 1 << irq_list[i];
    link->irq.Handler = NULL;
    
    /*
      General socket configuration defaults can go here.  In this
      client, we assume very little, and rely on the CIS for almost
      everything.  In most clients, many details (i.e., number, sizes,
      and attributes of IO windows) are fixed by the nature of the
      device, and can be hard-wired here.
    */
    link->conf.Attributes = 0;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;

    /* Allocate space for private device-specific data */
    local = kmalloc(sizeof(local_info_t), GFP_KERNEL);
    memset(local, 0, sizeof(local_info_t));
    link->priv = local;
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &ProxCSS_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    /* I got tired of figuring out how to get CardServices to give me
       an error - solved with a global.  :/  */
    if (LLSRetCSSConfig) {
	ProxCSS_cs_error(link->handle, RegisterClient, LLSRetCSSConfig);
    	link->open = 0;
    	link->state = 0;
	PCMCIACardInserted = CLEAR;
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    ((local_info_t *)link->priv)->stop = 1;
	    link->release.expires = jiffies + HZ/20;
	    add_timer(&link->release);
	}
	return link;
    }

    if ( register_netdev(&thisRangeLan) ) {
        /*    non-zero means failed init  */
	ProxCSS_cs_error(link->handle, RegisterClient, 1);
    	link->open = 0;
    	link->state = 0;
	PCMCIACardInserted = CLEAR;
	link->state &= ~DEV_PRESENT;
	((local_info_t *)link->priv)->stop = 1;
	link->release.expires = jiffies + HZ/20;
	add_timer(&link->release);
    }

    sprintf(link->dev->dev_name, thisRangeLan.name);

    return link;
} /* ProxCSS_attach */


/****************************************************************

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

    called by ProxCC_release

*****************************************************************/

static void ProxCSS_detach(dev_link_t *link)
{
    dev_link_t **linkp;

    DEBUG(0, "rlmod_detach(0x%p)\n", link);
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG)
	ProxCSS_release((u_long)link);

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if (link->priv) {
	kfree(link->priv);
    }
    kfree(link);
    
} /* ProxCSS_detach */


/****************************************************************

    ProxCSS_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.

    reads the tuples and sets up our variables.  This preceeds
    a call to ProxCSS_attach
    
*****************************************************************/

#define CS_CHECK(fn, args...) while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed
#define CFG_CHECK(fn, args...) if (CardServices(fn, args) != 0) goto next_entry

static void ProxCSS_config(dev_link_t *link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
    local_info_t *dev;
    int last_fn, last_ret;
    u_char buf[64];
    win_req_t req;
    memreq_t map;
    cistpl_mem_t *mem;
    
    handle = link->handle;
    dev = link->priv;

    DEBUG(0, "rlmod_config(0x%p)\n", link);

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* we use the manfid and card number to set flags.  Also
       we use them for some card config reporting in Prox_ioctl() */
    tuple.DesiredTuple = CISTPL_MANFID;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    LLSmanfid = parse.manfid.manf;
    LLSpid = parse.manfid.card;

    /* during product development, the product id isn't yet
       set but we can guess the card from the strings.  This
       should only be in here for me.  Production cards shouldn't
       need this.  */
    tuple.DesiredTuple = CISTPL_VERS_1;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);

    LLDFlag375 = SET;
    /* HomeRF only card */
    if ( LLSmanfid == PCCARD_VENDOR_ID_INTEL &&
	 !strcmp("HomeRF", parse.version_1.str+parse.version_1.ofs[0])) {
	wp = HOME_RF_ONLY;
    }
    /* non-HomeRF Proxim cards */
    if ( (LLSmanfid == PCCARD_VENDOR_ID_PROXIM) && 
	 !strcmp("PROXIM", parse.version_1.str+parse.version_1.ofs[0])) {
	LLDFlag375 = CLEAR;
    }
    
    /* Configure card */
    link->state |= DEV_CONFIG;

    /*
      In this loop, we scan the CIS for configuration table entries,
      each of which describes a valid card configuration, including
      voltage, IO window, memory window, and interrupt settings.

      We make no assumptions about the card to be configured: we use
      just the information available in the CIS.  In an ideal world,
      this would work for any PCMCIA card, but it requires a complete
      and accurate CIS.  In practice, a driver usually "knows" most of
      these things without consulting the CIS, and most client drivers
      will only use the CIS to fill in implementation-defined details.
    */
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    while (1) {
	cistpl_cftable_entry_t dflt = { 0 };
	cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
	CFG_CHECK(GetTupleData, handle, &tuple);
	CFG_CHECK(ParseTuple, handle, &tuple, &parse);

	if (cfg->index == 0) goto next_entry;
	link->conf.ConfigIndex = cfg->index;
	
	/* Use power settings for Vcc and Vpp if present */
	/*  Note that the CIS values need to be rescaled */
#ifndef SPOOF5V
	if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vcc = cfg->vcc.param[CISTPL_POWER_VNOM]/10000;
	else 
#endif
	     if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vcc = dflt.vcc.param[CISTPL_POWER_VNOM]/10000;

	if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vpp1 = link->conf.Vpp2 =
		cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
	else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vpp1 = link->conf.Vpp2 =
		dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
	
	/* Do we need to allocate an interrupt? */
	if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
	    link->conf.Attributes |= CONF_ENABLE_IRQ;
	
	/* IO window settings */
	link->io.NumPorts1 = link->io.NumPorts2 = 0;
	if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) {
	    cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
	/*  Some tuples are broken for this param so force IOCS16  
	    since all the cards use IOCS16 anyway */
	    link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
	    link->io.BasePort1 = io->win[0].base;
	    link->io.NumPorts1 = io->win[0].len;
	    if (io->nwin > 1) {
		link->io.Attributes2 = link->io.Attributes1;
		link->io.BasePort2 = io->win[1].base;
		link->io.NumPorts2 = io->win[1].len;
	    }
	}

	/* This reserves IO space but doesn't actually enable it */
	CFG_CHECK(RequestIO, link->handle, &link->io);

	/*
	  Note that the memory window base is a physical address, and
	  needs to be mapped to virtual space with ioremap() before it
	  is used.
	*/

	/* set up access to the attributes memory */
	  mem = (cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
	  req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_AM|WIN_ENABLE;
	  req.Base = 0;
	  req.Size = LLSMemorySize1;
	  req.AccessSpeed = 0;
	  link->win = (window_handle_t)link->handle;
	  CFG_CHECK(RequestWindow, &link->win, &req);
	  map.Page = 0; map.CardOffset = mem->win[0].card_addr;
	  CFG_CHECK(MapMemPage, link->win, &map);
/*  Contributed by Shinji Sumimoto */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,108)
      	req.Base = (long)ioremap(req.Base, 0x4000);
#endif
	/* If we got this far, we're cool! */
	break;
	
    next_entry:
	if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
	CS_CHECK(GetNextTuple, handle, &tuple);
    }
    
    /*
       Allocate an interrupt line.  Note that this does not assign a
       handler to the interrupt, unless the 'Handler' member of the
       irq structure is initialized.
    */
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	CS_CHECK(RequestIRQ, link->handle, &link->irq);
	
    /*
       This actually configures the PCMCIA socket -- setting up
       the I/O windows and the interrupt mapping, and putting the
       card and host interface into "Memory and IO" mode.
    */
    CS_CHECK(RequestConfiguration, link->handle, &link->conf);

    /*
      At this point, the dev_node_t structure(s) need to be
      initialized and arranged in a linked list at link->dev.
    */
    dev->node.major = dev->node.minor = 0;
    link->dev = &dev->node;

    /* Declare some values for Prox_probe */
    LLDIOAddress1 = link->io.BasePort1;
    LLDIntLine1 = link->irq.AssignedIRQ;
    LLDMemoryAddress1 = req.Base;

    /* Finally, report what we've done */
    printk(KERN_INFO "rlmod.o: index 0x%02x: Vcc %d.%d",
	   link->conf.ConfigIndex,
	   link->conf.Vcc/10, link->conf.Vcc%10);
    if (link->conf.Vpp1)
	printk(", Vpp %d.%d", link->conf.Vpp1/10, link->conf.Vpp1%10);
    if (link->win)
	printk(", mem 0x%06lx-0x%06lx", req.Base,
	       req.Base+req.Size-1);
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	printk(", irq %d", link->irq.AssignedIRQ);
    if (link->io.NumPorts1)
	printk(", io 0x%04x-0x%04x", link->io.BasePort1,
	       link->io.BasePort1+link->io.NumPorts1-1);
    if (link->io.NumPorts2)
	printk(" & 0x%04x-0x%04x", link->io.BasePort2,
	       link->io.BasePort2+link->io.NumPorts2-1);
    printk("\n");

    link->state &= ~DEV_CONFIG_PENDING;
    return;

cs_failed:
    ProxCSS_cs_error(link->handle, last_fn, last_ret);
    link->state &= ~DEV_CONFIG;
    LLSRetCSSConfig = CS_OUT_OF_RESOURCE;

} /* ProxCSS_config */


/****************************************************************

    After a card is removed, ProxCSS_release() will unregister the
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
    calls ProxCSS_detach and cleanup_module
*****************************************************************/

static void ProxCSS_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;

    DEBUG(0, "ProxCSS_release(0x%p)\n", link);

    /*
       If the device is currently in use, we won't release until it
       is actually closed, because until then, we can't be sure that
       no one will try to access the device or its data structures.
    */
    if (link->open) {
	DEBUG(1, "rlmod: release postponed, '%s' still open\n",
	      link->dev->dev_name);
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    /* Unlink the device chain */
    link->dev = NULL;

    /*
      In a normal driver, additional code may be needed to release
      other kernel data structures associated with this device. 
    */
    
    /* Don't bother checking to see if these succeed or not */
    CardServices(ReleaseConfiguration, link->handle);
    if (link->io.NumPorts1)
	CardServices(ReleaseIO, link->handle, &link->io);
    if (link->win)
	CardServices(ReleaseWindow, link->win);
    if (link->irq.AssignedIRQ)
	CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
	ProxCSS_detach(link);
    
} /* ProxCSS_release */


/****************************************************************

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.

    When a CARD_REMOVAL event is received, we immediately set a
    private flag to block future accesses to this device.  All the
    functions that actually access the device should check this flag
    to make sure the card is still present.
    
*****************************************************************/

static int ProxCSS_event(event_t event, int priority,
		       event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;

    DEBUG(1, "ProxCSS_event(0x%06x)\n", event);
    
    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
	PCMCIACardInserted = CLEAR;
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    ((local_info_t *)link->priv)->stop = 1;
	    link->release.expires = jiffies + HZ/20;
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	ProxCSS_config(link);
	break;
    case CS_EVENT_PM_SUSPEND:
	Prox_close(&thisRangeLan);
	link->state |= DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
	/* Mark the device as stopped, to block IO until later */
	((local_info_t *)link->priv)->stop = 1;
	if (link->state & DEV_CONFIG)
	    CardServices(ReleaseConfiguration, link->handle);
	break;
    case CS_EVENT_PM_RESUME:
	link->state &= ~DEV_SUSPEND;
	if (link->state & DEV_CONFIG)
	    CardServices(RequestConfiguration, link->handle, &link->conf);
	((local_info_t *)link->priv)->stop = 0;
	/*
	  In a normal driver, additional code may go here to restore
	  the device state and restart IO. 
	*/
	Prox_open(&thisRangeLan);
	break;
    case CS_EVENT_CARD_RESET:
	if (link->state & DEV_CONFIG)
	    CardServices(RequestConfiguration, link->handle, &link->conf);
	((local_info_t *)link->priv)->stop = 0;
	/*
	  In a normal driver, additional code may go here to restore
	  the device state and restart IO. 
	*/
	LLDNeedReset = SET;
	break;
    }
    return 0;
} /* ProxCSS_event */
#endif
